from wrapt import decorator

from inspect import getfullargspec
from copy import deepcopy
from typing import Generator, Callable, Any, Union, Type

from .constants import ErrorMessages


class Extension:
    @staticmethod
    def can_extend(target_cls: Type["Extendable"]) -> bool:
        """
        Should return a bool indicating whether this Extension can be applied to the target class
        """

        raise NotImplementedError

    @staticmethod
    def extend(target_cls: Type["Extendable"]) -> None:
        """
        Any modification of the target class should take place in this function
        """

        pass

    @staticmethod
    def _wrap(target_cls: Type["Extendable"], method_name: str,
              gen_func: Callable[["Extendable", Any, Any], Generator[None, Any, None]]) -> None:
        """
        Used to wrap an existing method on the target class.
        Passes copies of the method parameters to the generator function provided.
        The generator function should yield once,
        with the yield statement receiving a copy of the result of executing the core method
        """

        method = getattr(target_cls, method_name)
        method_args = getfullargspec(method).args

        if len(method_args) == 0 or method_args[0] != "self":
            ErrorMessages.wrap_static(method_name)

        @decorator  # This will preserve the original method signature when wrapping the method
        def wrapper(func, self, args, kwargs):
            gen = gen_func(self, *try_copy(args), **try_copy(kwargs))
            next(gen)

            result = func(*args, **kwargs)

            try:
                gen.send(try_copy(result))
            except StopIteration:
                pass

            return result

        setattr(target_cls, method_name, wrapper(method))

    @staticmethod
    def _set(target: Union[Type["Extendable"], "Extendable"], attribute_name: str, value: Any) -> None:
        """
        Used to safely add new attributes to an extendable class or instance. In contrast with assigning them directly,
        this method will raise an error if the attribute already exists (for example, if another extension added it)
        to ensure compatibility issues are flagged and can be dealt with easily
        """

        if hasattr(target, attribute_name):
            ErrorMessages.duplicate_attribute(attribute_name)

        setattr(target, attribute_name, value)


def try_copy(item: Any) -> Any:
    """
    A failsafe deepcopy wrapper
    """

    try:
        return deepcopy(item)

    except:
        return item



class ErrorMessages:
    @staticmethod
    def not_extension(extension):
        raise TypeError("A provided extension does not inherit the base Extension class: {}".format(extension))

    @staticmethod
    def invalid_extension(extension):
        raise ValueError("A provided extension cannot be used to extend this class: {}".format(extension))

    @staticmethod
    def wrap_static(property_name):
        raise ValueError(
            ("Static class methods cannot be wrapped. "
             "The method must receive the object instance as 'self' for its first argument: {}").format(property_name))

    @staticmethod
    def duplicate_attribute(attribute_name):
        raise AttributeError(
            "The provided attribute name already exists on the target instance: {}".format(attribute_name))



from typing import Type, FrozenSet

from .constants import ErrorMessages
from .extension import Extension


class Extendable:
    def __init__(self):
        self._extension_data = {}  # Intended to temporarily hold metadata - can be modified by extensions

    @classmethod
    def with_extensions(cls, *extensions: Type[Extension]) -> Type["Extendable"]:
        """
        Returns a copy of the class with the provided extensions applied to it
        """

        class Extended(cls):
            pass

        Extended._extensions = frozenset(extensions)

        for extension_cls in Extended._extensions:
            if not issubclass(extension_cls, Extension):
                ErrorMessages.not_extension(extension_cls)

            if not extension_cls.can_extend(cls):
                ErrorMessages.invalid_extension(extension_cls)

            extension_cls.extend(Extended)

        return Extended

    @property
    def extensions(self) -> FrozenSet[Type[Extension]]:
        return self._extensions
